// CForth  project, DMX related files  ****************************************
//
// Author and (c)2020-24 Wolfgang Schemmert. The copyright claim is restricted
// to this file and the depending files "main.c","main.h","cforth.c","cforth.h",
// "usb-otg-ctrlbd.c" and "globaldefines.h"
// Other depending material is published, copyrighted and licensed
// by STMicroelectronics and Segger GmbH/Rowley Associates Ltd
//
// Contact: <www.midi-and-more.de>
// A detailled hardware construction and operation manual
// is provided at this website
//
// ****************************************************************************
//      This program is free software: you can redistribute it and/or modify
//      it under the terms of the GNU General Public License as published by
//      the Free Software Foundation, either version 3 of the License, or
//      (at your option) any later version.
//
//      This program is distributed in the hope that it will be useful,
//      but WITHOUT ANY WARRANTY; without even the implied warranty of
//      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//      GNU General Public License for more details.
//
//      You should have received a copy of the GNU General Public License
//      along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//      Licence: GPL-3
//
// ****************************************************************************

#include "stm32f4xx.h"
#include "globaldefines.h"

// ************************************

//from main.h
extern uint32_t gLEDCOUNT, gTIMEOUT, gDELAYCOUNT,gLedBlink,gProject;
extern void sendTB(uint8_t);
extern uint32_t pullTermRxBuf(void);
extern void sendX8(uint8_t);
extern void sendX16(uint16_t);
extern void sendX32(uint32_t);

//from usb-fs-431.c
extern void sendUsbByte(uint8_t);

// ************************************

uint8_t gdMINIDMX_ON;  //gdNoZeroVelocity,

uint8_t gdDMXBUF[DMX_BUFLEN+1];     //actual DMX transmit buffer
uint8_t gdFINALMEM[DMX_BUFLEN+1];   //final level of fade process
uint8_t gdSTEPMEM[DMX_BUFLEN+1];    //used for control of fade timing
uint8_t gdSUBMEM[DMX_BUFLEN+1];
uint8_t gdRESIMEM[DMX_BUFLEN+1];
uint8_t gdDMXTEMP[DMX_BUFLEN+1];	//used by op_DMXFLASH

uint32_t gdDmxAddr, gdDmxTxIndex, gdDmxMax;
uint32_t gdFaderBusy, gdMiniDmxOn, gdMiniDmx;
uint32_t gdFadeTime, gdBootScene;

void startDmxTx(void);
void DmxHeader(void);
void wait16us(void);
void sendDmxByte(void);
void MiniDMX_Interpreter(void);
void MiniDMX_ACK(void);
void setDmxAddress(void);
void setDmxLevel(uint8_t);
void setNextDmxLevel(uint8_t);
void setEqual(uint8_t);
void fade2Preset(void);
void startFader(uint32_t);
void fadeProcess(void);
void readDmxBuf(void);
void sendAddrLevel(void);
//void sendLevel(uint16_t);

// ************************************

// DMX handling
//***************

void startDmxTx(void) // ****************************************************
// USART2-TxD is PortPA2, DMX-Break (Open Drain Pull Down) is PB0
// **************************************************************************
{
uint32_t k;

//USART2 is DMX transmitter with 250kBaud
	RCC->APB1ENR |=  RCC_APB1ENR_USART2EN ; // USART2 clock enable
	GPIOA->MODER &= ~(GPIO_MODER_MODER2_0 ); //Set PA2 to alternate function
	GPIOA->MODER |=   GPIO_MODER_MODER2_1 ;
	//GPIOA->OTYPER is push-pull by default
	GPIOA->OSPEEDR &= ~(GPIO_OSPEEDER_OSPEEDR2_1);
	GPIOA->OSPEEDR |= GPIO_OSPEEDER_OSPEEDR2_0;  //Medium Speed
//USART2 is AF7 on PortA
	GPIOA->AFR[0] &= 0xFFFFF0FF;	//PA2 is TX
	GPIOA->AFR[0] |= 0x00000700;
	USART2->BRR = 0x00A8; //250,000 Baud at 42MHz APB Clock (10.5=42M/250k/16)

// PB0 is Open Drain and pulls DMX OUT to SPACE during DMX Break=Reset
	//configure PB0 as open drain output, zero drive current
	GPIOB->OTYPER |= GPIO_OTYPER_OT_0;
	GPIOB->BSRR = GPIO_BSRR_BS_0;			//Open Drain High= transistor OFF!
	GPIOB->MODER &= ~GPIO_MODER_MODER0_1;
	GPIOB->MODER |= GPIO_MODER_MODER0_0;//PortA0 general purpose output mode
	GPIOB->OSPEEDR &= ~(GPIO_OSPEEDER_OSPEEDR0_1);
	GPIOB->OSPEEDR |= GPIO_OSPEEDER_OSPEEDR0_0;  //Medium Speed

	USART2->CR2 = USART_CR2_STOP_1;	//=2 Stopbits
	USART2->CR3 = 0;				//default:no Handshake
	USART2->CR1 = USART_CR1_UE | USART_CR1_TE;
									//TX only, 8Bit, no Parity, no Interrupts

	for (k=0;k < (DMX_BUFLEN+1);k++)  {
		gdDMXBUF[k] = 0;
		gdFINALMEM[k] = 0;
		gdSTEPMEM[k] = 0;
		gdSUBMEM[k] = 0;
		gdRESIMEM[k] = 0;
	}
	gdDmxAddr = 0;
//no interrupt is used
}


void DmxHeader(void)  // *****************************************************
// ***************************************************************************
{
uint8_t waitcount;

// Mark before Break min 0 max 1sec Includes transmission time of last DMX byte
	waitcount = 25;					//tested ca. 350 us
	while (waitcount)  {
		wait16us();
		waitcount-- ;
    }

// Space for Break = DMX loop Reset "typically" ca 176 us
// switch Open Drain output
	GPIOB->BSRR = GPIO_BSRR_BR_0;	//Open Drain Low= transistor ON !
 	waitcount = 18;					//tested ca 300us
	while (waitcount)  {
		wait16us();
		waitcount-- ;
	}
// release Open Drain output
    GPIOB->BSRR = GPIO_BSRR_BS_0;	//Open Drain High= transistor OFF !
// Mark after Break min 12 us
    wait16us();						//tested ca 35us
    wait16us();

// send DMX Startbyte = 0x00
	while ((USART2->SR & USART_SR_TXE) == 0) ;	//modified 13July20
	USART2->DR = 0;	//set DMX StartBit
// the TXE bit is cleared automatically when TDR is written
	gdDmxTxIndex = 0;               //index 0 of DMX buffer is User DMX addr=1
}

void wait16us(void)   // *******************************************************
// blocking delay ca 16us
// *************************************************************************
{
uint32_t countdown;

	countdown = 250;

	while (countdown)   {
		IWDG->KR = 0x0000AAAA;  //timeout ca. 1-2 seconds
		countdown-- ;
	}
}


void setDmxLevel(uint8_t level)  // *****************************************
// **************************************************************************
{
        if (gdFadeTime == 0)  {
        gdFINALMEM[gdDmxAddr] = level;
        gdDMXBUF[gdDmxAddr] = level;
    }
    else  {
        gdFINALMEM[gdDmxAddr] = level;
        startFader(gdDmxAddr);
    }
}


void startFader(uint32_t address)  // ***************************************
// **************************************************************************
{
uint32_t dividend, quotient;

    if (gdDMXBUF[address] == gdFINALMEM[address]) return;  //all-ready

    dividend = gdDMXBUF[address] - gdFINALMEM[address];
//calculate absolute difference,
    if (gdDMXBUF[address] > gdFINALMEM[address])
                            dividend = gdDMXBUF[address] - gdFINALMEM[address];
    else dividend =  gdFINALMEM[address] - gdDMXBUF[address];

    quotient = (dividend << 8) / gdFadeTime;
                                    //divisor != 0 guaranteed by program flow
    gdSTEPMEM[address] =(uint8_t)(quotient >> 8);    //HighByte
//gdSTEMPMEM contains the number of full gdDMXBUF steps each /100s  (fast fade)

    gdSUBMEM[address] =(uint8_t)(quotient & 0x000000FF); //LowByte
//because the the dividend was multiplied by 256, SUBMEM is a kind of remainder
//this is added to RESIMEM (residual of full step), ie. when RESIMEM overflows,
//a full extra +/-step (depends on fade direction) for DMXBUF gets inserted
//(slow fade), execution of this see fadeProcess().
    gdRESIMEM[address] = 0;
    gdFaderBusy = 1;
}


void fadeProcess(void)  // **************************************************
//this is the "ballistic" algorithm (target is computed at start time), with
//very small CPU overhead at runtime. It is less precise for long time fades
//with small differences between actual and final level.
//an alternative would be the "cruise missile" algorithm, which calculates the
//difference between actual and target every /100s and is more precise,
//but takes more CPU time (a division per DMX channel update).
// **************************************************************************
{
uint32_t k, residual;
int32_t actual, temp;
uint8_t step;

    gdFaderBusy = 0;                //temporary,check if further fade is active

    for (k=0;k<DMX_BUFLEN;k++)  {
        if (gdDMXBUF[k] == gdFINALMEM[k]) continue;
        gdFaderBusy = 0xFF;                 //trigger next fadeProcess()
        actual = (uint32_t)gdDMXBUF[k];
        step = gdSTEPMEM[k];
        residual = (uint32_t)gdRESIMEM[k];
        residual += gdSUBMEM[k];
        gdRESIMEM[k] = (uint8_t)(residual & 0x00FF);
        if (residual >= 0x0100) step++; //handle overflow without side effect
        if (actual < (uint32_t)gdFINALMEM[k])  {     //fade UP
            temp = (int32_t)(gdDMXBUF[k]) + step;
            if (temp > gdFINALMEM[k]) actual = gdFINALMEM[k];
            else actual += step;
        }
        else  {                                 //fade DOWN
            temp = (int32_t)(gdDMXBUF[k]) - step;
            if (temp < gdFINALMEM[k]) actual = gdFINALMEM[k];
            else  actual -= step;
        }
        gdDMXBUF[k] = (uint8_t)actual;
    }
}

// ##########################################################################



