// USB specific routines of of the CForth  project **************************
//
// Author and (c)2020-24 Wolfgang Schemmert. The copyright claim is restricted
// to this file and the depending files "main.c", "main.h", "cforth.c",
// "cforth.h", "dmx.c" and "globaldefines.h"
// Other depending material is published, copyrighted and licensed
// by STMicroelectronics and Segger GmbH/Rowley Associates Ltd
//
// Contact: <www.midi-and-more.de>
// A detailled hardware construction and operation manual 
// is provided at this website
//
// ****************************************************************************
//      This program is free software: you can redistribute it and/or modify
//      it under the terms of the GNU General Public License as published by
//      the Free Software Foundation, either version 3 of the License, or
//      (at your option) any later version.
//
//      This program is distributed in the hope that it will be useful,
//      but WITHOUT ANY WARRANTY; without even the implied warranty of
//      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//      GNU General Public License for more details.
//
//      You should have received a copy of the GNU General Public License
//      along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//      Licence: GPL-3
//
// ****************************************************************************


#include "stm32f4xx.h" //includes "stm32l476xx.h"
#include "globaldefines.h"

#define EP_MAXLEN  64
#define BUSPOWERED
#define PRODSTRINGLEN 24
#define MANUSTRINGLEN 18

const uint8_t CDC_DEVICEDESCRIPTOR [0x12] =
{
0x12,                   //length of this descriptor=18 bytes
1,                      //descriptor type 	DEVICE Descriptor = 1
//0x00,0x02,			//bdcUSB current rev of USB spec= 0x0110 = USB2.00
						//(host asks for HighSpeed mode, request not supported here.)
						//Therefore better take: (changed 05Oct21)
0x10,0x01,       		//bcdUSB current rev of USB spec= 0x0110 = USB1.10
2,                      //bDeviceClass	COMMUNICATION DEVICE CLASS = 2
0,       		        //bDeviceSubClass 	= 0x00 unused
0,                      //bDevice Protocol 	= 0x00, unused
EP_MAXLEN,             //bMaxPacketSize0 for endpoint 0
//if your device shall appear as "STMicroelectronics Virtual Com Port"
DEFAULT_VIDL,DEFAULT_VIDH, //idVendor VID= 0x0483
DEFAULT_PIDL,DEFAULT_PIDH, //idProduct PID= 0x5740
REVNUM,HARDWARE_VERSION,//bcd_Device	release code
0,                      //iManufacturer index of manufacturer string descriptor (unused)
0,   		            //iProduct	    index of product name string descriptor = 1
0,                      //iSerialNumber	serial number (unused)
1     	                //bNumConfigurations = 1 only one available
};


const uint8_t CDC_CONFIGDESCRIPTOR [67] =
{
9,                      //bLength		    length of this descriptor = 9 bytes
2,          	        //bDescriptor type	CONFIGURATION descriptor = 2

67,0,        			//wTotalLength	    total length of configuration block = 67 bytes
2,                      //bNumInterfaces    two interfaces incl this descriptor
1,      		        //bConfigurationValue ID of this configuration = 1
0,                      //iConfiguration	index of string descriptor:unused=0
#ifdef BUSPOWERED
0x80,                   //bmAttributes		BUS-powered - no remote wakeup
49,         	            //MaxPower		entry*2 = 98mA
#else  //default
0xC0,                   //bmAttributes		SELF-powered - no remote wakeup
0,                      //MaxPower 		    = 0 (self-powered only)
#endif

// CDC_INTERFACE_DESCRIPTOR
9,                      //bLength       	length of this descriptor = 9 bytes
4,                      //bDescriptorType	INTERFACE descriptor type=4
0,          	        //bInterfaceNumber	index of this interface =0
0,                      //bAlternateSetting	index of this setting =0
1,		                //bNumEndpoints 	(Notification endpoint) = 1
2,                      //bInterfaceClass 	= 2 (COMMUNICATION)
2,       		        //bInterfaceSubClass = 2 (ABSTRACT_CONTROL_MODEL)
0,                      //bInterfaceProtocol = 0 (unused)
0,        	            //iInterface 		= 0 (unused)

//COMMUNICATION DEVICE specific functional descriptor
5,                      //bLength			length of this descriptor = 5 bytes
0x24,       	        //bDescriptorType	CS_INTERFACEdescriptor type = 0x24
0,                      //bDescriptortSubtype HEADER subtype
0x10,1,                 //bcdADC	        =0x0100 revision of class specific descriptors

//CALL MANAGEMENT functional Descriptor
5,                      //bLength           length of this descriptor = 5 bytes
0x24,                   //bDescriptorType	CS_INTERFACEdescriptor type = 0x24
1, 			            //bDescriptorSubtype
3,                      //bmCapabilities	Device handles call management by itself
                        //ia DataInterface
						//i.e. in this case: no call management required
0,                     	//bDataInterface:	Data interface no.0 (see below) associated
//1,                     	//bDataInterface:	Data interface no.0 (see below) associated

//ABSTRACT CONTROL management functional Descriptor
4,                      //bLength			length of this descriptor = 4 bytes
0x24,         	        //bDesriptorType	CS_INTERFACEdescriptor type = 0x24
2,                      //bDescriptorSubtype
7,				        //bmCapabilities    Send_Break,Set_Line_Coding... etc

//UNION functional descriptor
5,                      //bLength			length of this descriptor = 5 bytes
0x24,                   //bDescriptorType	CS_INTERFACE descriptor
6,                      //bDescriptorSubtype UNION FUNCTIONAL descriptor
0,		       	        //bMasterInterface
1,                      //bSlaveInterface0

//Notification standard endpoint descriptor
7,  			        //bLength           length of this descriptor = 7 bytes
5,                      //bDescriptorType	ENDPOINT
0x83,			        //bEndpointAddress	no.3, IN
3,                      //bmAttributes 	INTERRUPT = 3
0x10,0,                 //wMaxPacketSize	=16
64,			            //interval in ms    =64

//DATA CLASS INTERFACE descriptor
9,                      //bLength			length of this descriptor = 9 bytes
4,				        //bDescriptorType	INTERFACE descriptor type
1,                      //bInterfaceNumber
0,				        //bAlternateSetting
2,                      //bNumEndpoints
0x0A,       			//bInterfaceClass	Data Class
0,                      //bInterfaceSubclass unused
0,				        //bInterfaceProtocol no specific protocol
0,                      //iInterface 		unused

//DATA BULK IN endpoint standard descriptor
7,				        //bLength           length of this descriptor = 7 bytes
5,                      //bDescriptorType	ENDPOINT
0x81,			        //bEndpointAddress	nr.1, IN
2,                      //bmAttributes		=2 (bulk)
0x40,0,			        //wMaxPacketSize	=64
1,				        //bInterval			=1ms

//DATA BULK OUT endpoint standard descriptor
7,                      //bLength			length of this descriptor = 7 bytes
5,				        //bDescriptorType	ENDPOINT
2,                      //bEndpointAddress	nr.=2, OUT
2,				        //bmAttributes 		=2 (bulk)
0x40,0,			        //wMaxPacketSize	=64
1				        //bInterval			=1ms
};

//#ifdef STLINK_CUTOFF
//asked by "Raspbian"
const uint8_t COUNTRY[4] = 
{
4,3,9,4						//inserts country code 0409 in other string req
};

//Manufacturer String Descriptor  index 2
const uint8_t MANUFACT_STRINGDESCRIPTOR [MANUSTRINGLEN] =
{
MANUSTRINGLEN,			//bLength = 16+2  Descriptor length in Bytes
0x03,                   //bDescriptorType = 0x03, Constant String = 3
'h',0,'o',0,'m',0,'e',0,'m',0,'a',0,'d',0,'e',0
};
//#endif

//Product String Descriptor index 1
const uint8_t PRODUCT_STRINGDESCRIPTOR [PRODSTRINGLEN] =
{
PRODSTRINGLEN,			//bLength = 22+2  Descriptor length in Bytes
0x03,                   //bDescriptorType = 0x03, Constant String = 3
'C',0,'F',0,'o',0,'r',0,'t',0,'h',0,'-',0,'M',0,'I',0,'D',0,'I',0
                        //bString (Unicode 16Bit)
};

const uint8_t LINE_CODING_115200 [7] =
{
0x00,0xC2,0x01,0x00,0,0,8  //0x0001C200=decimal 115200, rest describes 1N8
//the baud rate is faked, the microcontroller accepts any and sets its own for the USART.
//but 115200 seems to be the best default at host side. Else enter your baudrate 32bit formatted.
};

const uint8_t DEVICE_STATUS[2] = 
{
0,0						//Bus powered, no Remote Wakep
};

const uint8_t MIDI_DEVICEDESCRIPTOR[0x12] =
{
18,        	            //bLength			length of this descriptor=18 bytes
1,                     	//descriptor type 	DEVICE Descriptor = 1
0x00,0x02,              //bdcUSB current rev of USB spec= 0x0110 = USB2.00
0,      		        //bDeviceClass		Device defined at Interface level
0,                     	//bDeviceSubClass 	= 0x00 unused
0,	    	            //bDevice Protocol 	= 0x00, unused
EP_MAXLEN,	            //bMaxPacketSize0	Endp.0 = 64
//if your device shall appear as "STMicroelectronics Virtual Com Port"
DEFAULT_VIDL,DEFAULT_VIDH, //idVendor       DEFAULT_VID= 0x0483
DEFAULT_PIDL+MIDI_PLUS,DEFAULT_PIDH, 
						//idProduct DEFAULT_PID=0x5743! else host problem!
//DEFAULT_PIDL,DEFAULT_PIDH, //idProduct DEFAULT_PID=0x5742! else host problem!
REVNUM,HARDWARE_VERSION,//bcd_Device release code
//#ifdef STLINK_CUTOFF
2,   		            //Manufacturer	index of manufacturer string descriptor
//#else
//0,   		            //Manufacturer	index of manufacturer string descriptor
//#endif
1,                   	//iProduct Index of product name string descriptor
0,     		            //iSerialNumber		serial number (unused)
1                    	//bNumConfigurations = 1 only one available
};

const uint8_t MIDI_CONFIGDESCRIPTOR[0x65] =  //dez101
{
9,                      //bLength length of this descriptor = 9 bytes
2,                      //bDescriptorType CONFIGURATION descriptor type = 2
0x65,0,			        //wTotalLength length of configuration block =101
                        // incl this descriptor
2,      		        //bNumInterfaces  two interfaces
1,						//bConfigurationValue  ID of this configuration
0,                      //iConfiguration  index of string descriptor:unused=0
#ifdef BUSPOWERED
0x80,                    //bmAttributes		BUS-powered - no remote wakeup
49,                      //MaxPower 		entry*2 = 98mA
#else
0xC0,                   //bmAttributes		SELF-powered - no remote wakeup
0,                      //MaxPower			= 0 (self-powered only)
#endif
//Standard Audio Class Interface Descriptor
9,                      //bLength			length of this descriptor = 9 bytes
4,          	        //bDescriptorType	INTERFACE descriptor type=4
0,                      //bInterfaceNumber	index of this interface
0,	                    //bAlternativeSetting index of this setting
0,	                    //bNumEndpoints		= 0 (EP0 only)
1,       		        //bInterfaceClass	= 0x01 (AUDIO)
1,                      //bInterfaceSubClass = 0x01 (AUDIO_CONTROL)
0,         	            //bInterfaceProtocol = 0 (unused)
0,                      //iInterface 		= 0 (unused)
//Class-specific AudioControl Interface descriptor
9,       	            //bLength			length of this descriptor = 9 bytes
0x24,                   //bDescriptorTye	CS_INTERFACEdescriptor type = 0x24
1,       		        //bDescriptorSubtyte HEADER subtype
0,1,                    //bcdADC			=0x0100 revision of class specific descriptors
9,0,                    //wTotalLength		= 9
1, 			            //bInCollection		number of streaming interfaces
1,   			        //baInterfaceNr(1)	MIDIStreaming interface no.1
						//					belongs to this Audio Control
//Standard MIDI Streaming Interface Descriptor
9,                      //bLength	        length of this descriptor = 9 bytes
4,         	            //bDescriptorType	INTERFACE descriptor type = 0x04
1,                      //ibInterfaceNumber	index of this interface
0,				        //bAlternativeSetting index of this setting
2,						//bNumENdpoints		2 endpoints
1, 			            //bInterfaceClass	AUDIO
3, 						//bInterfaceSubclass MIDISTREAMING
0,          	        //bInterfaceProtocol = 0 (unused)
0,                       //iInterface 		= 0 (unused)
//Class specific MIDI Streaming Interface Descriptor
7,                      //bLength 			length of this descriptor = 7 bytes
0x24,                   //bDescriptorType	CD_INTERFACE descriptor
1,       	            //bDescriptorSubtye	MS_HEADER subtype
0,0x01,                 //bcdADC			Revision of this class specification 0x0100
0x41,0x00,		        //wTotalLength		total size of class-specific descriptors
//MIDI IN Jack descriptor (embedded)
6,                      //bLength           length of this descriptor = 6 bytes
0x24,			        //bDescriptorType	CS_INTERFACE descriptor
2,						//bDescriptorSubtype MIDI_IN_JACK subtype
1,				        //bJackType			EMBEDDED
1,						//bJackID			ID of this jack
0,				        //iJack 			unused
//MIDI IN Jack descriptor (external)
6,			            //bLength           length of this descriptor = 6 bytes
0x24,			        //bDescriptorType	CS_INTERFACE descriptor
2,						//bDescriptorSubtype MIDI_IN_JACK subtype
2,				        //bJackType			EXTERNAL
2, 						//bJackID			ID of this jack
0,				        //iJack 			unused
//MIDI OUT Jack descriptor (embedded)
9,			            //bLength           length of this descriptor = 9 bytes
0x24,			        //bDescriptorType	CS_INTERFACE descriptor
3, 						//bDescriptorSubtype MIDI_OUT_JACK subtype
1,				        //bJackType			EMBEDDED
3, 						//bJackID			ID of this jack
1,				        //bNrInputPins		= 1
2,						//BaSourceID(1)		this jack is connected to jack with ID 2
1,				        //BaSourcePin(1)	there connected to pin 1
0,   					//iJack 			unused
//MIDI OUT Jack descriptor (external)
9,			            //length of this descriptor = 9 bytes
0x24,					//bDescriptorType	CS_INTERFACE descriptor
3,				        //bDescriptorSubtype MIDI_OUT_JACK subtype
2, 						//bJackType			EXTERNAL
4,				        //bJackID			ID of this jack
1,						//bNrInputPins		= 1
1,				        //BaSourceID(1)		this jack is connected to jack with ID 1
1,						//BaSourcePin(1)	there connected to pin 1
0,  			        //iJack 			unused
//Standard BULK IN endpoint descriptor
9,						//bLength           length of this descriptor = 9 bytes
5,			            //bDescriptorType	ENDPOINT descriptor
0x81,					//bEndpointAddress	IN endpoint no.1
2,			            //bmAttributes		BULK DATA, not shared
0x40,0x00,				//wMaxPacketSize	64 bytes/packet
0,			            //bInterval 		ignored for Bulk = 0
0,				        //bRefresh 			unused
0,						//bSynchAddress 	unused
//Class-specific BULK IN endpoint descriptor
5,			            //bLength			length of this descriptor = 5 bytes
0x25,					//bDescriptorType	CS_ENDPOINT descriptor
1,				        //bDescriptorSubtype MS_GENERAL subtype
1,						//bNumEmbMIDIJack	= 1 number of embedded MIDI OUT jacks
3,				        //BaAssocJackID(1)	associated with embedded MIDI OUT no.3
//Standard BULK OUT endpoint descriptor
9,						//bLength           length of this descriptor = 9 bytes
5,				        //bDescriptorType	ENDPOINT descriptor
2,   					//bEndpointAddress	OUT endpoint no.2
2,			            //bmAttributes		BULK DATA,not shared
0x40,0x00,				//wMaxPacketSize	64 bytes/packet
0,			            //bInterval 		ignored for Bulk = 0
0,				        //bRefresh 			unused
0,						//bSynchAddress 	unused
//Class-specific BULK OUT endpoint descriptor
5,                      //bLength			length of this descriptor = 5 bytes
0x25,			        //bDescriptorType	CS_ENDPOINT descriptor
1,				        //bDescriptorSubtype MS_GENERAL subtype
1,						//bNumEmbMIDIJack	= 1 number of embedded MIDI IN jacks
1				        //baAssocJackID(1)	associated with embedded MIDI IN no.1
};


// *****************************************************

extern uint32_t gVid, gPid, gSplit;

extern void LED_Off(void);
extern void wait5us(void);
extern void msDelay(uint16_t) ;
extern void servSystem(void);
//USART
extern void sendB(uint8_t);
extern void sendTB(uint8_t);
extern void sendAllB(uint8_t);
extern void sendX8(uint8_t);
extern void sendX32(uint32_t);

//from cforth.c
extern uint32_t gfTermCtrlBD;	//message from terminal input for backgnd proc
extern void op_DEBUG(void);

// ******************************************************

uint8_t guTempConfigBuf[0x12];

uint8_t gUSB_TXBUF[USB_BUFLEN+2]; //for EP1 IN
uint8_t *gUSB_TXHEAD=0,*gUSB_TXTAIL=0,*gUSB_TXBUFTOP=&(gUSB_TXBUF[USB_BUFLEN]);
uint8_t gUSB_RXBUF[USB_BUFLEN+2]; //for EP2 OUT
uint8_t *gUSB_RXHEAD=0,*gUSB_RXTAIL=0,*gUSB_RXBUFTOP=&(gUSB_RXBUF[USB_BUFLEN]);

extern uint32_t   gTest;      //used to check system variables without SVD

//USB Std Request parameters
uint8_t    guRequestType;
uint8_t    guRequest;
uint16_t   guValue;
uint16_t   guIndex;
uint16_t   guLength;

//USB packet handling
uint8_t    guAddress, guDescriptor, guDescIndex;
uint16_t   guTx0RemainCnt, guTx1RemainLen;
uint32_t guDeviceClass, guConfigured, guLockEP2, guSetupAction;
const uint8_t *guTx0SendPtr;
uint16_t    guTx1Len;  //actual fill state of gUSB_TXBUF to be sent at EP1
uint8_t     guRx2Len;  //remaining number of bytes to be received at EP2

//MIDI special
uint8_t	guMSTATUSok, guMDATA1ok, guMDATA2ok, guSYSEXBODY, guTxPackCnt;
uint8_t guMSTATUS,guMIDIDAT1, guMIDIDAT2, guPACKTYPE;
uint8_t guRx2RunState;   //running state of EP2 MIDI Messages

void stopUSB(void);
void initUSB(void);
void handle_USB_Reset(void);

void USB_OTG_IRQHandler(void);
void readRxFifo(void);
void handle_USB_EpOut(void);
void handle_USB_EpIn(void);
void handle_SetupStage(void);
void sendDataEp0_IN(void);
void sendZLP0(void);
void stallEP0_OUT(void);
void sendDataEp1_IN(void);

void sendUsbByte(uint8_t);
uint32_t pullUsbRxBuf(void);
uint8_t deCodeMidi(uint32_t);
void enCodeMidi(uint8_t);

void stopUSB(void)   // *****************************************************
//  ***************************************************************************
{
    GPIOA->AFR[1] &= 0xFFF00FFF;                //de-configure GPIOA 11,12 as AF
    GPIOA->MODER&=~(GPIO_MODER_MODER11|GPIO_MODER_MODER12); //now input
    NVIC_DisableIRQ(OTG_FS_IRQn);
    RCC->AHB2RSTR |= RCC_AHB2RSTR_OTGFSRST;     //reset USB-OTG
//    USB->CNTR |= USB_CNTR_PDWN;                 //USB is disabled completely
    RCC->AHB2ENR &= ~RCC_AHB2ENR_OTGFSEN;       //disable USB-OTG clock
    guConfigured = 0;
}


void initUSB(void)   // *****************************************************
// PortA Clock already active
// Main Clock already: HSE, AHB=24MHz, APB1=24MHz, APB2=24MHz
// USB clock already set to 48 MHz
//  ***************************************************************************
{

    RCC->AHB2RSTR |= RCC_AHB2RSTR_OTGFSRST;     //reset USB-OTG
    RCC->AHB2RSTR &= ~RCC_AHB2RSTR_OTGFSRST;    //and release reset
    RCC->AHB2ENR &= ~RCC_AHB2ENR_OTGFSEN;       //disable USB-OTG clock


//enable USB power (provided VDDUSB is connected by externally)
// SB31, SB57 and SB60 are closed on the Nucleo board
//    PWR->CR2 |= PWR_CR2_USV; //as precaution if VDDUSB is not connected externally

//    sendX32(*OTG_PCGCCTL);

//PA11 is USB_FS_DM
//PA12 is USB_FS_DP
//USB_FS_SOF and USB_FS_ID not needed
//PA11,12 --> AF10=0xA
    GPIOA->AFR[1] &= 0xFFF00FFF;
    GPIOA->AFR[1] |= 0x000AA000;
//configure GPIOA 11,12 as AF
	GPIOA->MODER &= ~(GPIO_MODER_MODER11_0|GPIO_MODER_MODER12_0);
	GPIOA->MODER |=  (GPIO_MODER_MODER11_1|GPIO_MODER_MODER12_1);

//set GPIOA 11,12 Speed up to 25MHz
    GPIOA->OSPEEDR &= 0xFC3FFFFF ;              //OSPEEDR 11,12_0+1 down
    GPIOA->OSPEEDR |= 0x01400000 ;              //OSPEEDR 11,12_0 up
//output push-pull type is reset state
//set GPIOA input type to HiZ
    GPIOA->PUPDR &= 0xFC3FFFFF;                 //PUPDR11,12_0+1 down

    RCC->AHB2ENR |= RCC_AHB2ENR_OTGFSEN;        //enable USB-OTG clock


//"DCD_Init()":
    OTG_FS->GAHBCFG &= 0xFFFFFFFE;          //disable GINTMASK
    OTG_FS->DCFG = 3;                   //init Full Speed, set address = 0
//"USB_OTG_CoreInit()":
//perform Core Soft Reset
//"USB_OTG_CoreReset()":
    while ((OTG_FS->GRSTCTL & 0x80000000) == 0) ;
                                //wait until AHBIDL bit cleared (AHB bus idle)

    OTG_FS->GRSTCTL |= 1; //OTG_FS_GRSTCTL_CSRST//perform Core Soft Reset
    msDelay(1);
    while ((OTG_FS->GRSTCTL & 0x80000001) != 0x80000000) ;
    msDelay(1);

    OTG_FS->DCTL &=  ~mask1;        //clear SDIS ="soft disconnect" flag
    msDelay(2);

//configuration of USB_OTG_FS Core registers:

	OTG_FS->GCCFG =    mask21 | mask16;	//enable NOVBUSBSENS (don't check PA9)
										//deactivate Power Down
#ifdef STLINK_CUTOFF
//check Vbus presence by software
//reconfirm PA9 (Vbus sensing) as HiZ input
	GPIOA->MODER &= ~(GPIO_MODER_MODER9);
    GPIOA->PUPDR &= 0xFFF3FFFF;                 //PUPDR9_0+1 down
//Alternative with slightly modified hard-and firmware
//(USART1_TX=PB6, PA9 connected with VBus). Preferably for self-powered devices
//    OTG_FS->GCCFG = (OTG_GCCFG_VBDEN | OTG_GCCFG_PWRDWN);
                        //enable Vbus Sensing and deactivate Power Down
#endif
    msDelay(50);
    OTG_FS->GUSBCFG |= (1<<30|          //set FDMOD = force Device Mode
                        6<<10|          //set TRDT
                        1<<6|           //set PHYSEL
                        0x10);			//set no TOCAL
    msDelay(50);

//"OTG_EnableCommonInt()":
//"USB_OTG_EnableCommonInt()":
    OTG_FS->GINTSTS = 0xFFFFFFFF;       //clear all pending core interrupts
    OTG_FS->GINTMSK  = (mask31|			//Wakeup Int
                        mask13|			//Enumeration Done Int
                        mask12|			//USB Reset Int
                        mask11|			//USB Suspend Int
                        mask3);			//SOF Int

//configuration of USB_FS_OTG device registers
//"USB_OTG_CoreInitDev()":
    OTG_FS->PCGCCTL = 0;                //restart PHY clock

    OTG_FS->DIEPMSK = 0;
    OTG_FS->DOEPMSK = 0;
    OTG_FS->DAINTMSK = 0;

    handle_USB_Reset();

    while (!(OTG_FS->GINTSTS & ~(mask13|mask12)));
                                        //wait for ENUMDNE and USBRST Int
//wait for ENUMDNE interrupt and check USB_OTG->DSTS if USB speed is correct
    msDelay(30);
    OTG_FS->GINTSTS |= mask13|mask12;         //clear these flags

//"USB_OTG_EP0Activate()":
//MPSIZ in DIEPCTL0 already set
   OTG_FS->DCTL |= mask8;                 //clear CGINAK (global IN NAK)

    NVIC_EnableIRQ(OTG_FS_IRQn);
}


void handle_USB_Reset(void)    // *****************************************
// *************************************************************************
{

    guConfigured = 0;

    OTG_FS->DCFG = 3;			//init Full Speed, set address = 0

//setup data FIFO RAM for each of the FIFOs
    OTG_FS->GRXFSIZ = 128;	//=STM USB library RX_FIFO_FS_SIZE
     OTG_FS->DIEPTXF0=(64<<16)|128;	//len=TX0_FIFO_FS_SIZE,startaddr=RX_FIFO_FS_SIZE
     OTG_FS->DIEPTXF1 = 64<<16|256;//IN EP =STM USB library TX1_FIFO_FS_SIZE start=0
     OTG_FS->DIEPTXF2 = 0;			//OUT EP=STM USB library TX2_FIFO_FS_SIZE
    if (guDeviceClass == CDC_CLASS)  OTG_FS->DIEPTXF3 = 32<<16|384;//*OTG_DIEPTXF3 = 32<<16|384;
								//IN EP=STM USB library TX3_FIFO_FS_SIZE-start=0

//flush the FIFOs
    OTG_FS->GRSTCTL |= mask5|mask4; //set TXFFLSH and RXFFLSH
    while ((OTG_FS->GRSTCTL & 0x00000030) != 0x00000000) ;

//Device endpoint initialization on USB reset:
    OTG_FS->DIEPCTL0 = 0;			//max. packet size = 64
    OTG_FS->DIEPINT0 |= 0x0000005B;
    OTG_FS->DIEPTSIZ0 = 0;
    OTG_FS->DOEPCTL0 = mask31;         //max. packet size = 64
    OTG_FS->DOEPINT0 |= 0x0000005B;
    OTG_FS->DOEPTSIZ0 = (3<<29)|(1<<19)|64;//STUPCNT = 3,PKTCNT=1,XFRSIZ=64

	OTG_FS->DIEPCTL1 = (1<<22)|(2<<18)|64;
								//FIFO#1, Bulk IN endpoint, max packet size=64
    OTG_FS->DIEPINT1 |= 0x0000005B;
    OTG_FS->DIEPTSIZ1 = 0;
    OTG_FS->DOEPCTL1 = mask22;		//formally Bulk IN endpoint, no MPSIZ
    OTG_FS->DOEPINT1 |= 0x0000005B;
    OTG_FS->DOEPTSIZ1 = 0;

    OTG_FS->DIEPCTL2 = (2<<22);     //formally FIFO#2, Bulk OUT endpoint, no MPSIZ
    OTG_FS->DIEPINT2 |= 0x0000005B;
    OTG_FS->DIEPTSIZ2 = 0;
    OTG_FS->DOEPCTL2 = (2<<22)|(2<<18)|64; //Bulk OUT endpoint, max packet size=64
    OTG_FS->DOEPINT2 |= 0x0000005B;
    OTG_FS->DOEPTSIZ2 = (1<<19)|64;

    if (guDeviceClass == CDC_CLASS)  {
        OTG_FS->DIEPCTL3 = (3<<22)|(3<<18)|16;
							//FIFO#3, Interrupt IN endpoint, max packet size=16
        OTG_FS->DIEPINT3 |= 0x0000005B;
        OTG_FS->DIEPTSIZ3 = 0;
        OTG_FS->DOEPCTL3 = (3<<22);     //formally Interrupt IN endpoint, no MPSIZ
        OTG_FS->DOEPINT3 |= 0x0000005B;
        OTG_FS->DOEPTSIZ3 = 0;
    }
    else  {  //MIDI
        OTG_FS->DIEPCTL3 = 0;               //FIFO#3 closed
        OTG_FS->DIEPINT3 |= 0x0000005B;
        OTG_FS->DIEPTSIZ3 = 0;
        OTG_FS->DOEPCTL3 = 0;
        OTG_FS->DOEPINT3 |= 0x0000005B;
        OTG_FS->DOEPTSIZ3 = 0;
    }

    OTG_FS->DOEPMSK |= mask3|1;            //XFRCM and STUPM unmasked
    OTG_FS->DIEPMSK |= mask4|mask3|1;     //ITTXFEMSK, TOM and XFRCM unmasked

//"USB_OTG_EnableDevInt()":
//first disable all interrupts
    OTG_FS->GINTMSK = 0;

//perform "USB_OTG_EnableCommonInt()" again with additional interrupt types:
	OTG_FS->GINTSTS = 0xFFFFFFFF;   //clear all pending core interrupts
	OTG_FS->GINTMSK |= (mask19|        //OUT interrupts unmasked
                        mask18|        //IN interrupts unmasked
                        mask13|        //Enumeration Done Int
                        mask12|        //USB Reset Int
						mask11|        //USB Suspend Int
						mask4|        //RXFLVL Int
						mask3);       //SOF Int

//"USB_OTG_EnableGlobalInt()":
   OTG_FS->GAHBCFG |= mask7|1;         //set TXFELVL and GINTMASK bit

//see usb_core.c - USB_OTG_EPActivate():
   OTG_FS->DAINTMSK =  (mask18|        //EP2(OUT) EP interrupt unmasked
						mask16|        //EP0(OUT) EP interrupt unmasked
						mask1|        //EP1(IN) EP interrupt unmasked
						1);        //EP0(IN) EP interrupt unmasked
    if (guDeviceClass == CDC_CLASS)  OTG_FS->DAINTMSK |= 1<<3;
                                    //EP3(IN) EP interrupt unmasked

    OTG_FS->DIEPEMPMSK = 0;

// EP0 is always activated by the core in all configurations !!!

    guSetupAction = 0;
    guTx0RemainCnt = 0;

    gUSB_TXHEAD = gUSB_TXBUF;   //flush EP1 application level buffer
    gUSB_TXTAIL = gUSB_TXBUF;
    gUSB_RXHEAD = gUSB_RXBUF;   //flush EP2 application level buffer
    gUSB_RXTAIL = gUSB_RXBUF;
    guTxPackCnt = 0;            //prepare enCodeMidi()
    guRx2RunState = 0;          //prepare deCodeMidi()
}


void OTG_FS_IRQHandler(void)    // *****************************************
// *************************************************************************
{
uint32_t  GINT_Status;

    GINT_Status = OTG_FS->GINTSTS & OTG_FS->GINTMSK;
    if (GINT_Status == 0) return;                   //spurios interrupt

	if (GINT_Status & mask19)  handle_USB_EpOut();			//OEPINT

    if (GINT_Status & mask18)  handle_USB_EpIn();			//IEPINT

    if (GINT_Status & mask4)   readRxFifo();				//RXFLVL flag

    if (GINT_Status & mask3)   {
		OTG_FS->GINTSTS  = mask3;						//clear SOF flag
    }
    if (GINT_Status & mask12)  {
        handle_USB_Reset();
        OTG_FS->GINTSTS  = mask12;						//clear RESET flag
    }
    if (GINT_Status & mask13)  OTG_FS->GINTSTS = mask13;	//clear ENUMDONE flag
    if (GINT_Status & mask31)  OTG_FS->GINTSTS = mask31;	//clear WKUINT flag
    if (GINT_Status & mask11)  OTG_FS->GINTSTS = mask11;	// clear USBSUSP flag
}


void readRxFifo(void) // ***************************************************
// *************************************************************************
{
uint32_t * fifo;
volatile uint32_t status, value1, value2, fillstate;
uint8_t nextbyte;

   OTG_FS->GINTMSK &= ~mask4;                 //mask bit4:RXFLVLM

    fifo = (uint32_t *) OTG_FS_DFIFO0_BASE;
    status = OTG_FS->GRXSTSP;
    value1 = status & 0x001E7FFF;

    if ((status & 0x0000000F) == 0) {   //-->EP0f
        if (value1 == 0x00040000) { //filter PKTSTS=DATA OUT,BCNT=0->ZLP packet
        OTG_FS->GINTMSK |= mask4;          //unmask RXFLVLM, enable Int. again
		return;	//25Jan24
        }
        if (value1 == 0x00040070) { //filter SET_LINE_CODING packet, 7 bytes
			wait5us();
			wait5us();
            value1 = (*(volatile uint32_t *)(fifo));
            value2 = (*(volatile uint32_t *)(fifo));
//            guSetupAction = 0x20;
            guTx0RemainCnt = 0;
			sendZLP0();				//25Jan24
        }
        if ((status & 0x001E0000) == 0x000C0000) {  //SETUP packet, always 8 bytes
            value1 = (*(volatile uint32_t *)(fifo));
            value2 = (*(volatile uint32_t *)(fifo));
//parse immediately
            guRequestType = (uint8_t) (value1 & 0x000000FF);
            guRequest = (uint8_t) ((value1>>8) & 0x000000FF);
            guValue = (uint16_t) ((value1>>16) & 0x0000FFFF);
            guDescriptor = (uint8_t) ((guValue >> 8) & 0x00FF);
			guDescIndex = (uint8_t)(guValue & 0x00FF);
            guIndex = (uint16_t) (value2 & 0x0000FFFF);
            guLength = (uint16_t) ((value2>>16) & 0x0000FFFF);
        }
    }

    if ((status & 0x001E000F) == 0x00040002) {
//EP2 OUT packet recveived  *****************
//-->EP2 is BULK OUT
        guRx2Len = (uint8_t) (((status & 0x00000FF0)>>4) & 0x000000FF); //(<=64

//restore endpoint buffer into application level buffer
        while ((guRx2Len > (guDeviceClass*3)) && (guRx2Len < 65 )) {
                                    // limits are protection against underflow
            value1 = (*(volatile uint32_t *)(fifo));
            if (guDeviceClass == MIDI_CLASS)  {
//transform USB MIDI packets into MIDI byte stream
//regular MIDI packets always consist of 4 bytes
                deCodeMidi(value1) ;
            }
            else  {  //CDC_CLASS
//restore immediately into cyclic buffer
                nextbyte = (uint8_t) ((value1 & 0x000000FF) & 0xFF);
//modification 11June22 and 26Feb24. 
//Formally needed for compiler/linker, not really used
//#ifndef ORIG_NUCLEO
				if ((gSplit!=2)&&(nextbyte==CTRL_B))  gfTermCtrlBD = CTRL_B;
				else if ((gSplit!=2)&&(nextbyte==CTRL_D)) gfTermCtrlBD=CTRL_D;
//#else
//				if (nextbyte == CTRL_B) gfTermCtrlBD = CTRL_B;
//				else if (nextbyte == CTRL_D) gfTermCtrlBD = CTRL_D;
//#endif
				else  {
					*gUSB_RXHEAD++ = nextbyte ;
					if (gUSB_RXHEAD == gUSB_RXBUFTOP)  gUSB_RXHEAD=gUSB_RXBUF;
				}
                guRx2Len-- ;
                if (guRx2Len == 0) break;
				
                nextbyte = (uint8_t) (((value1 & 0x0000FF00)>>8) & 0xFF);
//#ifndef ORIG_NUCLEO
				if ((gSplit!=2)&&(nextbyte==CTRL_B))  gfTermCtrlBD = CTRL_B;
				else if ((gSplit!=2)&&(nextbyte==CTRL_D)) gfTermCtrlBD=CTRL_D;
//#else
//				if (nextbyte == CTRL_B) gfTermCtrlBD = CTRL_B;
//				else if (nextbyte == CTRL_D) gfTermCtrlBD = CTRL_D;
//#endif
				else  {
					*gUSB_RXHEAD++ = nextbyte ;
					if (gUSB_RXHEAD == gUSB_RXBUFTOP)  gUSB_RXHEAD=gUSB_RXBUF;
				}
                guRx2Len-- ;
                if (guRx2Len == 0) break;
				
                nextbyte = (uint8_t) (((value1 & 0x00FF0000)>>16) & 0xFF);
//#ifndef ORIG_NUCLEO
				if ((gSplit!=2)&&(nextbyte==CTRL_B))  gfTermCtrlBD = CTRL_B;
				else if ((gSplit!=2)&&(nextbyte==CTRL_D)) gfTermCtrlBD=CTRL_D;
//#else
//				if (nextbyte == CTRL_B) gfTermCtrlBD = CTRL_B;
//				else if (nextbyte == CTRL_D) gfTermCtrlBD = CTRL_D;
//#endif
				else  {
					*gUSB_RXHEAD++ = nextbyte ;
					if (gUSB_RXHEAD == gUSB_RXBUFTOP)  gUSB_RXHEAD=gUSB_RXBUF;
				}
                guRx2Len-- ;
                if (guRx2Len == 0) break;
				
                nextbyte = (uint8_t) (((value1 & 0xFF000000)>>24) & 0xFF);
//#ifndef ORIG_NUCLEO
				if ((gSplit!=2)&&(nextbyte==CTRL_B))  gfTermCtrlBD = CTRL_B;
				else if ((gSplit!=2)&&(nextbyte==CTRL_D)) gfTermCtrlBD=CTRL_D;
//#else
//				if (nextbyte == CTRL_B) gfTermCtrlBD = CTRL_B;
//				else if (nextbyte == CTRL_D) gfTermCtrlBD = CTRL_D;
//#endif
				else  {
					*gUSB_RXHEAD++ = nextbyte ;
					if (gUSB_RXHEAD == gUSB_RXBUFTOP)  gUSB_RXHEAD=gUSB_RXBUF;
				}
                guRx2Len-- ;
                if (guRx2Len == 0) break;
            }
        }

//calculate fill level of gUSB_RXBUF
//TAIL>HEAD?
        if (gUSB_RXTAIL > gUSB_RXHEAD) {
            fillstate = (uint32_t)((gUSB_RXHEAD + USB_BUFLEN) - gUSB_RXTAIL) ;
        }
        else fillstate = (uint32_t)(gUSB_RXHEAD - gUSB_RXTAIL) ;

 // lock EP2 interrupt when gUSB_RXBUF contains at least (SIZE-64) bytes
                    // gets unlocked in routine pullUsbRxBuf
                    // as soon as gUSB_RXBUF contains less than (SIZE-64) bytes
                    //this way EP2 always accidently can accept
                    //another packet of 64 bytes without data loss
                    //(similar to CTS/RTS handshake at RS-232)
        guLockEP2 = 0;
        if (fillstate > USB_BUFLEN - 70)  guLockEP2 = 0xFF;
    }
        OTG_FS->GINTMSK |= mask4;          //unmask RXFLVLM, enable Int. again
}


void handle_USB_EpIn(void)    // *******************************************
// *************************************************************************
{
volatile uint8_t  DAINT_EP;
volatile uint32_t temp;

    DAINT_EP =
	   (uint8_t)((OTG_FS->DAINT & OTG_FS->DAINTMSK)&0x000000FF);
												//"USB_OTG_ReadDevAllInEpItr()"

    if (DAINT_EP & 0x01)  {
//EP0 IN packet  *********

// if bit0 (TXFE) is set in DIEPEMPMSK,
// then set bit7 in temp as additional mask with DIEPINT0
        temp = OTG_FS->DIEPMSK | ((OTG_FS->DIEPEMPMSK & 1)<<7);
        temp = OTG_FS->DIEPINT0 & temp;
                                    //see "usb_dcd-int.c-->DCD_ReadDevInEP()"
        if (temp & 0x00000080)  {
//TXFE=TX fifo empty flag in DIEPINT0 set

            OTG_FS->DIEPEMPMSK &= 0xFFFFFFFE;
                                    //clear TXFE flag for Ep0(DIEPEMPMSK/bit0)
            if ((guSetupAction==1)||(guSetupAction==6)||(guSetupAction==0xA1)) {
                sendDataEp0_IN();
            }
        }

        if (temp & 0x00000001)  {
//XFERC flag in DIEPINT0 set
            OTG_FS->DIEPINT0  = 0x00000001;     //clear XFERC Complete flag
            if (guTx0RemainCnt == 0)   {
                if ((guSetupAction==5)||(guSetupAction==9)||
					(guSetupAction==0x22) || (guSetupAction==0x20)) {
                    if (guConfigured == 0) { //last chance, should already be configured
                        OTG_FS->DIEPCTL1 = mask15|mask31|mask26|(1<<22)|(2<<18)|64;
                                    // activate EP,enable EP,clear NAK,
                                    //FIFO#1,Bulk IN endpoint,max packetsize=64
                        OTG_FS->DOEPCTL2 = mask15|mask31|mask26|(2<<22)|(2<<18)|64;
                                // activate EP,enable EP,clear NAK,
                                //FIFO#2,Bulk OUT endpoint, max packetsize=64
                        if (guDeviceClass == CDC_CLASS)  {
                        OTG_FS->DIEPCTL3 = mask15|mask31|mask26|(3<<22)|(3<<18)|16;
                                // activate EP,enable EP,clear NAK,
                                //FIFO#3,Interrupt IN endpoint,max packetsize=16
                        }
                    }
                }

                OTG_FS->DIEPEMPMSK &= 0xFFFFFFFE;
                                    //disable TXFE int. for Ep0(DIEPEMPMSK/bit0)
//prepare for next expected Setup packet:
                OTG_FS->DOEPTSIZ0 = (3<<29)|(1<<19)|64;
                                    //STUPCNT = 3,PKTCNT=1,XFRSIZ=64
                OTG_FS->DOEPCTL0 |= mask31|mask26|0;
                                    //set EPENA,clear NAK,max.packsize=64
            }
            else   {
//prepare for next following IN packet
                OTG_FS->DIEPTSIZ0 = (1<<19) | guTx0RemainCnt;
                                                // PKTCNT|bytes_to_send
                OTG_FS->DIEPCTL0 |= mask31|mask26|0;
                                        //set EPENA,clear NAK,max.packsize=64
                OTG_FS->DIEPEMPMSK |= 1; //set TXFE flag for Ep0(DIEPEMPMSK/bit0)
                              //i.e.trigger remaining packet(s) of the Descriptor
            }
        }
        if (temp & mask4) OTG_FS->DIEPINT0 = mask4;     //clear ITTXFEMSK bit
        if (temp & mask3) OTG_FS->DIEPINT0 = mask3;     //clear TOC bit
    }

    if (DAINT_EP & 0x02)  {
//EP1 IN packet ******************************

// if bit1 is set in DIEPEMPMSK, 
//then set bit7 in temp as combined mask for DIEPINT1
        temp =OTG_FS->DIEPMSK | ((OTG_FS->DIEPEMPMSK & 2)<<6);
        temp = OTG_FS->DIEPINT1 & temp;
        if (temp & 0x00000080)  {
//TXFE=TX fifo empty flag DIEPINT1 set

            OTG_FS->DIEPEMPMSK &= 0xFFFFFFFD;
                            //disable TXFE int. for Ep1(DIEPEMPMSK/bit1)
            sendDataEp1_IN();           //fill FIFO
        }

        if ((temp & 0x00000001) || (temp & mask4)) {
//XFERC flag in DIEPINT1 set
//or ITTXFEMSK bit set: IN request while FIFO empty
            if (temp & 0x00000001) OTG_FS->DIEPINT1 = 0x00000001;
                                                //clear XFER Complete flag

            if (gUSB_TXTAIL == gUSB_TXHEAD) {   //i.e. transferSize = 0
                OTG_FS->DIEPCTL1 |= mask31|mask26;    // enable EP, clear NAK
                OTG_FS->DIEPEMPMSK &= 0xFFFFFFFD;
                                //disable TXFE int. for Ep1(DIEPEMPMSK/bit1)
            }
            else  { //if (guTx1Len == 0) {
                    //calculate fill level of gUSB_TXBUF
                if (gUSB_TXTAIL > gUSB_TXHEAD) {  //if TAIL>HEAD:abs += BUFLEN
                  guTx1Len = (uint16_t)((gUSB_TXHEAD+USB_BUFLEN)-gUSB_TXTAIL);
                }
                else guTx1Len = (uint16_t)(gUSB_TXHEAD - gUSB_TXTAIL);
                if (guTx1Len > EP_MAXLEN) guTx1Len = EP_MAXLEN;
                OTG_FS->DIEPTSIZ1 = (1<<19)|guTx1Len;   // PKTCNT=1,+XFERSIZ
                OTG_FS->DIEPCTL1 |= mask31|mask26;		// enable EP, clear NAK
                OTG_FS->DIEPEMPMSK |= 0x00000002;
                                    //enable TXFE int. for Ep1(DIEPEMPMSK/bit1)
            }

            if (temp & mask4) OTG_FS->DIEPINT1 = mask4; //clear ITTXFEMSK bit
        }

        if (temp & mask3) OTG_FS->DIEPINT1 = mask3; //Timeout: clear TOC bit
    }


    if (DAINT_EP & 0x08)  {
//EP3 IN packet  *******************************

// if bit3 is set in DIEPEMPMSK,
//then set bit7 in temp as combined mask for DIEPINT3
        temp = OTG_FS->DIEPMSK | ((OTG_FS->DIEPEMPMSK & 8)<<4);
        temp = OTG_FS->DIEPINT3 & temp;

        if (temp & 0x00000080)  {
//TXFE=TX fifo empty flag in DIEPINT3 set
            OTG_FS->DIEPTSIZ3 = (1<<19);			// PKTCNT=1,XFERSIZ =0
            OTG_FS->DIEPCTL3 |= mask31|mask26;		//enable EP,clear NAK
        }

        if (temp & 0x00000001)  {
//XFERC flag in DIEPINT3 set
//by default no notification messages: always return ZLP
            OTG_FS->DIEPTSIZ3 = (1<<19);            // PKTCNT=1,XFERSIZ =0
            OTG_FS->DIEPCTL3 |= mask31|mask26;    //enable EP,clear NAK
            OTG_FS->DIEPINT3  = 0x00000001;         //clear XFER Complete flag
            OTG_FS->DIEPEMPMSK &= 0xFFFFFFF7;
                                    //clear TXFE flag for Ep3(DIEPEMPMSK/bit3)
        }

        if (temp & mask4){                     //IN request while FIFO empty
//start or keep data IN process running: initiate ZLP packet
            OTG_FS->DIEPTSIZ3 = (1<<19);            // PKTCNT=1,XFERSIZ = 0
            OTG_FS->DIEPCTL3 |= mask31|mask26;    // enable EP, clear NAK
            OTG_FS->DIEPEMPMSK &= 0xFFFFFFF7;
                                    //clear TXFE flag for Ep3(DIEPEMPMSK/bit3)
            OTG_FS->DIEPINT3 = mask4; //clear ITTXFEMSK bit
         }
        if (temp & mask3) OTG_FS->DIEPINT3 = mask3; //Timeout: clear TOC bit
    }
}


void handle_USB_EpOut(void)    // *****************************************
// *************************************************************************
{
volatile uint8_t  DAINT_EP, temp;

    DAINT_EP = (uint8_t)(((OTG_FS->DAINT & OTG_FS->DAINTMSK) >> 16) & 0x000000FF);

    if (DAINT_EP & 0x01)  {                      //EP0 OUT packet
//*****************************
        temp = (uint8_t) (OTG_FS->DOEPINT0 & OTG_FS->DOEPMSK & 0x000000FF);
        if (temp & 0x01)   {
//XFERComplete flag
            OTG_FS->DOEPINT0  = 0x00000001;         //clear XFER Complete flag
            OTG_FS->DOEPTSIZ0 = (3<<29)|(1<<19)|64;
											//STUPCNT = 3,PKTCNT=1,XFRSIZ=64
            OTG_FS->DOEPCTL0 |= mask31|mask26|0;
									//set EPENA,clear NAK,max.packet size=64
//HACK!!  ##################
			if ((guRequestType == 0x80) && (guRequest == 0))  temp |= 0x08;
		// #############
			else  {
				if (guSetupAction == 0x20)  {
											//"SetLineCoding" send ZLP packet
					OTG_FS->DIEPTSIZ0 = (1<<19);    //PKTCNT|len=0
					OTG_FS->DIEPCTL0 |= mask31|mask26|0;
									//set EPENA,clear NAK,max.packet size=64
				}
			}
        }
        if (temp & 0x08) {
//STUP flag, SETUP packet
            handle_SetupStage();            //-->action + bufferAddress + length
            if (guSetupAction==5)  {
//"DCD_EP_SetAddress":
                OTG_FS->DCFG |= (guAddress<<4);         //insert new address
                sendZLP0();
/*
				OTG_FS->DIEPTSIZ0 = (1<<19);            // len=0|PKTCNT
                OTG_FS->DIEPCTL0 |= m31|m26|0;
                                        //set EPENA,clear NAK,max.packet size=64
                OTG_FS->DOEPINT0  = 0x00000008;         //clear SETUP DONE flag
                                          //ZLP packets don't use TXFE interrupt
                OTG_FS->>DIEPEMPMSK |= 1;  //set TXFE flag for Ep0(DIEPEMPMSK/bit0)
*/
            }

            if ((guSetupAction==9) || (guSetupAction == 0x22)) {
//"SetConfiguration" or "setLineState":
//send ZLP packet, nothing else to be done
                OTG_FS->DIEPTSIZ0 = (1<<19);            //PKTCNT|len=0
                OTG_FS->DIEPCTL0 |= mask31|mask26|0;
                                        //set EPENA,clear NAK,max.packet size=64
                OTG_FS->DOEPINT0  = 0x00000008;         //clear SETUP DONE flag
                                          //ZLP packets don't use TXFE interrupt
                OTG_FS->DIEPEMPMSK |= 1; //set TXFE flag for Ep0(DIEPEMPMSK/bit0)
            }

            if ((guSetupAction==1)||(guSetupAction==6)||(guSetupAction==0xA1)) {
                if (guTx0RemainCnt>EP_MAXLEN) OTG_FS->DIEPTSIZ0 = mask19|EP_MAXLEN;
                else OTG_FS->DIEPTSIZ0 = (1<<19) | guTx0RemainCnt;
														//PKTCNT|bytes_to_send
                OTG_FS->DIEPCTL0 |= mask31|mask26|0;
                                        //set EPENA,clear NAK,max.packet size=64
                OTG_FS->DOEPINT0 = 0x00000008;//clear SETUP DONE flag
                OTG_FS->DIEPEMPMSK |= 1; //set TXFE flag for Ep0(DIEPEMPMSK/bit0)
            }
        }
        return;
    }

    if (DAINT_EP & 0x04)  {
//EP2 OUT packet ********

        temp = (uint8_t) (OTG_FS->DOEPINT2 & OTG_FS->DOEPMSK & 0x000000FF);
        if (temp & 0x01)  {
//XFERComplete flag
            OTG_FS->DOEPINT2  = 0x00000001; //clear XFER Complete flag
            OTG_FS->DOEPTSIZ2 = (1<<19)|64; //PKTCNT=1,XFRSIZ=64:int each packet

            if (guLockEP2==0) OTG_FS->DOEPCTL2 |= mask31|mask26; //set EPENA,clear NAK
            else OTG_FS->DOEPCTL2 |= mask31; //set EPENA, DON'T clear NAK
        }
    }
}


void handle_SetupStage(void)    // ******************************************
// **************************************************************************
{
uint16_t k;

    guSetupAction = 0;

    if (guRequestType == 0x21)  {   //exclusively sent for CDC_CLASS
                    // special case SET_LINE_CODING or SET_COMTROL_LINE_STATE
        if (guRequest == 0x20)  {

            OTG_FS->DOEPINT0 |= mask3;     //clear STUP flag
            OTG_FS->DOEPTSIZ0 = (3<<29)|(1<<19)|64;
                                        //STUPCNT = 3,PKTCNT=1,XFRSIZ=64
            OTG_FS->DOEPCTL0 |= mask31|mask26|0;
                                        //set EPENA,clear NAK,max.packet size=64
            guSetupAction = 0x20;
        }
        else if (guRequest == 0x22) {
            guSetupAction = 0x22;
            guTx0RemainCnt = 0;
        }
        else  {
            stallEP0_OUT();
            return;
        }
    }
    else if((guRequestType==0xA1)&&(guRequest==0x21)) { //exclusively CDC_CLASS
                                                //special case GET_LINE_CODING
        guSetupAction = 0xA1;
        guTx0RemainCnt = 7;
        if (guLength < guTx0RemainCnt) guTx0RemainCnt = guLength;
        guTx0SendPtr = LINE_CODING_115200;
    }
    else if (((guRequestType & 0x61) == 0) || (( guRequestType & 0x61) == 1)) {
        switch(guRequest) {

            case 0:                             //Get_Status
				guSetupAction = 1; //to be destinguished from "noAction"
                guTx0RemainCnt = 2;
				if (guLength < guTx0RemainCnt) guTx0RemainCnt = guLength;
                guTx0SendPtr = DEVICE_STATUS;
                break;

            case 0x05:                          //Set_Address
                guAddress = (uint8_t) (guValue & 0x007F);
                                                //keep new address globally
                guSetupAction = 5;
                guTx0RemainCnt = 0;
                break;

            case 0x06:                          //Get_Descriptor
                guSetupAction = 6;
                if (guDescriptor == 1)  {
                    guTx0RemainCnt = 0x12;       //18
					if (guLength < guTx0RemainCnt) guTx0RemainCnt = guLength;
					if (guDeviceClass == MIDI_CLASS) {
                        guTx0SendPtr = MIDI_DEVICEDESCRIPTOR;
                    }
                    else  guTx0SendPtr = CDC_DEVICEDESCRIPTOR;
//if configured individually, assemble user defined VID/PID

                    if ((gVid != DEFAULT_VID) || (gPid != DEFAULT_PID)) {
                                 //Vid/Pid in flash  is modified by user
						for (k=0;k<0x12;k++)  {
							guTempConfigBuf[k] = guTx0SendPtr[k];
						}
                        guTempConfigBuf[8] =
                                (uint8_t)(gVid & 0x000000FF);
                        guTempConfigBuf[9] =
                                (uint8_t)((gVid>>8) & 0x000000FF);
                       
						if (guTx0SendPtr == MIDI_DEVICEDESCRIPTOR)  {
							guTempConfigBuf[10] =
								(uint8_t)((gPid & 0x000000FF) + MIDI_PLUS);
						}
						else  {
							guTempConfigBuf[10] =
										(uint8_t)(gPid & 0x000000FF);
						}
						guTempConfigBuf[11] =
                                (uint8_t)((gPid>>8) & 0x000000FF);

						guTx0SendPtr = guTempConfigBuf;
					}
				}
                else if (guDescriptor == 2)  {
                    if (guDeviceClass == MIDI_CLASS) {
                        guTx0RemainCnt = 0x65;       //101
                        guTx0SendPtr = MIDI_CONFIGDESCRIPTOR;
                    }
                    else  {
                        guTx0RemainCnt = 0x43;       //67
                        guTx0SendPtr = CDC_CONFIGDESCRIPTOR;
                    }
                    if (guLength < guTx0RemainCnt) guTx0RemainCnt = guLength;
                }
                else if (guDescriptor == 3)  {
                    if (guDescIndex == 1)  {
						guTx0RemainCnt = PRODSTRINGLEN;
						if (guLength <guTx0RemainCnt) guTx0RemainCnt=guLength;
						guTx0SendPtr = PRODUCT_STRINGDESCRIPTOR;
					}
//#ifdef STLINK_CUTOFF
                    else if (guDescIndex == 2)  {
						guTx0RemainCnt = MANUSTRINGLEN;
						if (guLength <guTx0RemainCnt) guTx0RemainCnt=guLength;
						guTx0SendPtr = MANUFACT_STRINGDESCRIPTOR;
					}
                    else if (guDescIndex == 0)  {
						guTx0RemainCnt =4;
						if (guLength <guTx0RemainCnt) guTx0RemainCnt=guLength;
						guTx0SendPtr = COUNTRY;
					}
//#endif
					else  {
						stallEP0_OUT();
						return;
					}
                }
                else stallEP0_OUT();
                break;

            case 0x09:                          //Set_Configuration
                guSetupAction = 9;
                guTx0RemainCnt = 0;              //send ZLP
                guConfigured = 0xFF;
                OTG_FS->DIEPCTL1 = mask15|mask31|mask26|(1<<22)|(2<<18)|64;
                                    // activate EP,enable EP,clear NAK,
                                    //FIFO#1,Bulk IN endpoint,max packetsize=64
                OTG_FS->DOEPCTL2 = mask15|mask31|mask26|(2<<22)|(2<<18)|64;
                                // activate EP,enable EP,clear NAK,
                                //FIFO#2,Bulk OUT endpoint, max packetsize=64
                if (guDeviceClass == CDC_CLASS)  {
                   OTG_FS->DIEPCTL3 = mask15|mask31|mask26|(3<<22)|(3<<18)|16;
                                // activate EP,enable EP,clear NAK,
                                //FIFO#3,Interrupt IN endpoint,max packetsize=16

                }
                break;
            default:
                stallEP0_OUT();
                break;
        }
    }
    else stallEP0_OUT();
}


void sendZLP0(void)    // **************************************************
// send EP0 ZLP packet to Host
// *************************************************************************
{
//send ZLP packet, nothing else to be done

	OTG_FS->DIEPTSIZ0 = (1<<19);	//PKTCNT|len=0
	OTG_FS->DIEPCTL0 |= mask31|mask26;	//set EPENA,clear NAK,max.packet size=64
	OTG_FS->DOEPINT0  = 0x00000008;	//clear SETUP DONE flag
									//ZLP packets don't use TXFE interrupt
	OTG_FS->DIEPEMPMSK |= 1; 	//set TXFE flag for Ep0(DIEPEMPMSK/bit0)
}


void stallEP0_OUT(void)    // **************************************************
// **************************************************************************
{
//    if (*OTG_DIEPCTL0 & m31)  *OTG_DIEPCTL0 |= m30;
                                            //first disable EP IN  ?????????
    OTG_FS->DOEPCTL0 |= mask21;        //set STALL bit, will be cleared by core
//    *OTG_DOEPTSIZ0 = (3<<29)|(1<<19)|64; //STUPCNT = 3,PKTCNT=1,XFRSIZ=64
                                    //enable EP0 for setup packets
}


void sendDataEp0_IN(void)    // *****************************************
// prepare FIFO for DATA IN stage
// *************************************************************************
{
uint32_t * fifo;
const uint8_t * descPtr;
volatile uint32_t littleEndianWord;
volatile uint8_t len, len32b, k;


    if (guTx0RemainCnt > EP_MAXLEN)  {
        descPtr = guTx0SendPtr;
        guTx0SendPtr += EP_MAXLEN;
        len = EP_MAXLEN;
        guTx0RemainCnt -= EP_MAXLEN;
    }
    else {
        descPtr = guTx0SendPtr;
        len = guTx0RemainCnt;
        guTx0RemainCnt = 0;
    }

        len32b = (len + 3)/4;
        fifo = (uint32_t *) OTG_FS_DFIFO0_BASE;

    for (k=0;k<len32b;k++)  {
        littleEndianWord =  (uint32_t)(*(descPtr+3)<<24) |
                            (uint32_t)(*(descPtr+2)<<16) |
                            (uint32_t)(*(descPtr+1)<<8) |
                            (uint32_t)(*descPtr) ;
        *fifo = littleEndianWord;
        descPtr += 4;
    }
}


void sendDataEp1_IN(void)    // *********************************************
// **************************************************************************
{
uint32_t * fifo;
volatile uint32_t littleEndianWord;
volatile uint32_t temp0, temp1, temp2, temp3;

    fifo = (uint32_t *) OTG_FS_DFIFO1_BASE;
    littleEndianWord = OTG_FS->DTXFSTS1;        //must be read formally
    while (guTx1Len)  {
        temp0 = (uint32_t)*gUSB_TXTAIL;
        gUSB_TXTAIL++;
        if (gUSB_TXTAIL == gUSB_TXBUFTOP)  gUSB_TXTAIL = gUSB_TXBUF;
        guTx1Len-- ;
        temp1 = 0;
        temp2 = 0;
        temp3 = 0;

        if (guTx1Len) {
            temp1 = (uint32_t)*gUSB_TXTAIL;
            gUSB_TXTAIL++;
            if (gUSB_TXTAIL == gUSB_TXBUFTOP)  gUSB_TXTAIL = gUSB_TXBUF;
            guTx1Len-- ;
        }
        if (guTx1Len) {
            temp2 = (uint32_t)*gUSB_TXTAIL;
            gUSB_TXTAIL++;
            if (gUSB_TXTAIL == gUSB_TXBUFTOP)  gUSB_TXTAIL = gUSB_TXBUF;
            guTx1Len-- ;
        }
        if  (guTx1Len) {
            temp3 = (uint32_t)*gUSB_TXTAIL;
            gUSB_TXTAIL++;
            if (gUSB_TXTAIL == gUSB_TXBUFTOP)  gUSB_TXTAIL = gUSB_TXBUF;
            guTx1Len-- ;
        }
        littleEndianWord =  (temp3<<24) |
                            (temp2<<16) |
                            (temp1<<8) |
                             temp0 ;
        *fifo = littleEndianWord;

    }
//without a delay here, the communication tends to break down.
//Exact reason unknown.
//this hack takes least CPU time in the interrupt routine:
//    temp0 = 600;
//    while (temp0) temp0--;
   while (OTG_FS->DTXFSTS1 != EP_MAXLEN) {
 //          sdbgX32(*OTG_DTXFSTS1);
   }
}

// ###########################################################################
// Application level USB interface
//********************************

uint32_t pullUsbRxBuf(void) // **********************************************
// **************************************************************************
{
volatile uint32_t recv, fillstate;

    servSystem();   //serve watchdog etc.

    if (guConfigured == 0)  return 0;

// check first if something received. If yes, then set Bit15. If not, return 0
    if (gUSB_RXTAIL != gUSB_RXHEAD)  {
        recv= (uint32_t)(*gUSB_RXTAIL++ & 0x00FF);
        if (gUSB_RXTAIL == gUSB_RXBUFTOP)  gUSB_RXTAIL = gUSB_RXBUF;

//calculate fill level of gUSB_RXBUF
//TAIL>HEAD?
        if (gUSB_RXTAIL > gUSB_RXHEAD) {
            fillstate =(uint32_t)gUSB_RXHEAD+USB_BUFLEN-(uint32_t)gUSB_RXTAIL;
        }
        else fillstate = (uint32_t)gUSB_RXHEAD - (uint32_t)gUSB_RXTAIL ;

// unlock EP2 when EP2_BUF contains less than (SIZE-64) bytes
        if (fillstate < (USB_BUFLEN - 70)) {
            guLockEP2 = 0;
            OTG_FS->DOEPCTL2 |= mask31|mask26; //set EPENA,clear NAK
        }
		
		LED_Off();
		return (recv | 0x00008000);

// removed 11June22
/*
		if ((gSplit != 2) && (gSplit != 4)) {
		//USB is terminal and MiniDmx is not active
		//particularly ignore CTRL_D if USB is serial I/O
//			if (recv == CTRL_B) gfTermCtrlBD = CTRL_B;	
			if (recv != CTRL_D) return (recv | 0x00008000);
			else  {
				op_DEBUG();		//execute immediately
				return 0;		//don't forward CTRL_D
			}
		}
		else  return (recv | 0x00008000);	//ignore CTRL_D in any case
*/
    }
	else return 0;   // tail = head
}


uint8_t deCodeMidi(uint32_t midipack)   // **********************************
//decode MIDISTREAMING packets to USART MIDI byte stream
//may only be called from "readRxFifo" interrupt routine!
// **************************************************************************
{
uint8_t index, status, data1, data2;

//regular MIDI packets always consist of 4 bytes:
    index = (uint8_t) ((midipack & 0x000000FF) & 0xFF);
    guRx2Len-- ;
    if (guRx2Len == 0)  return 0;
    status = (uint8_t) (((midipack & 0x0000FF00)>>8)& 0xFF);
    guRx2Len-- ;
    if (guRx2Len == 0)  return 0;
    data1 = (uint8_t) (((midipack & 0x00FF0000)>>16) & 0xFF);
    guRx2Len-- ;
    if (guRx2Len == 0)  return 0;
    data2 = (uint8_t) (((midipack & 0xFF000000)>>24) & 0xFF);
    guRx2Len-- ;
    if (index > 0x0F)  return 0;
            //wrong cable number, does not correspond with embedded MIDI IN jack
    if (index < 2)  return 0; //reserved for future, seems transmission error

	LED_Off();

    if ((index == 5) || (index == 0x0F))  {
                //1 byte System Common, SysEx or undefined content
#ifdef RUNNINGSTATE
        guRx2RunState = 0;			//invalidate running state
#endif
        *gUSB_RXHEAD++ = status ;   //send one byte in any case
        if (gUSB_RXHEAD == gUSB_RXBUFTOP)  gUSB_RXHEAD=gUSB_RXBUF;
        return 1;
    }
    if ((index == 0x0C)  || (index == 0x0D))  {
                            //2 bytes MIDI Channel Message
#ifdef RUNNINGSTATE
        if (guRx2RunState != status)  {
            *gUSB_RXHEAD++ = status ;
            if (gUSB_RXHEAD == gUSB_RXBUFTOP)  gUSB_RXHEAD=gUSB_RXBUF;
            guRx2RunState = status;
        }
#else
        *gUSB_RXHEAD++ = status ;
        if (gUSB_RXHEAD == gUSB_RXBUFTOP)  gUSB_RXHEAD=gUSB_RXBUF;
#endif

        *gUSB_RXHEAD++ = data1 ;
        if (gUSB_RXHEAD == gUSB_RXBUFTOP)  gUSB_RXHEAD=gUSB_RXBUF;
        return 1;
	}
	if ((index == 2)  || (index == 6))  {
                            //2 bytes System Common or SysEX content
        guRx2RunState = 0;
        *gUSB_RXHEAD++ = status ;
        if (gUSB_RXHEAD == gUSB_RXBUFTOP)  gUSB_RXHEAD=gUSB_RXBUF;
        *gUSB_RXHEAD++ = data1 ;
        if (gUSB_RXHEAD == gUSB_RXBUFTOP)  gUSB_RXHEAD=gUSB_RXBUF;
        return 1;
	}

//all yet unfiltered packets have 3 bytes content:

#ifdef RUNNINGSTATE
    if (index < 8)  {
                    //don't use running state with Realtime Messages
        guRx2RunState = 0;
        *gUSB_RXHEAD++ = status ;
        if (gUSB_RXHEAD == gUSB_RXBUFTOP)  gUSB_RXHEAD=gUSB_RXBUF;
    }
    else  {         //Channel Message
        if (guRx2RunState != status)  {
            *gUSB_RXHEAD++ = status ;
            if (gUSB_RXHEAD == gUSB_RXBUFTOP)  gUSB_RXHEAD=gUSB_RXBUF;
            guRx2RunState = status;
        }
    }
#else
    *gUSB_RXHEAD++ = status ;
    if (gUSB_RXHEAD == gUSB_RXBUFTOP)  gUSB_RXHEAD=gUSB_RXBUF;
#endif

    *gUSB_RXHEAD++ = data1 ;
    if (gUSB_RXHEAD == gUSB_RXBUFTOP)  gUSB_RXHEAD=gUSB_RXBUF;
    *gUSB_RXHEAD++ = data2 ;
    if (gUSB_RXHEAD == gUSB_RXBUFTOP)  gUSB_RXHEAD=gUSB_RXBUF;
    return 1;
}


void sendUsbByte(uint8_t byte)  // ******************************************
// add data byte to EP1 transmit buffer
// **************************************************************************
{
    if (guConfigured == 0)  return;
    *gUSB_TXHEAD++ = byte;
    if (gUSB_TXHEAD == gUSB_TXBUFTOP)  gUSB_TXHEAD=gUSB_TXBUF;
}


void enCodeMidi(uint8_t recv)   // *******************************************
//prepare USART MIDI byte stream for transmission as USB MIDISTREAMING packets
//take one byte per loop
// **************************************************************************
{
uint8_t msgType;
uint8_t *tempHead;

    guTxPackCnt++ ;

    if (recv & 0x80)   {   //recv is status byte
        if (recv != EOX)  guTxPackCnt = 1; //chance to resync corrupt byte stream
        if (recv == 0xF0)   {
//handle beginning of SysEx message
//special case:no running state,but has a data body which looks like running state
            guMSTATUS = 0xF0;
            guMSTATUSok = 0xFF;
            guMDATA1ok = 0;
            guMDATA2ok = 0;
            guSYSEXBODY = 0;
            return;			//wait for subsequent bytes
        }
//collect 1-3 bytes before packet is sent
        if (recv == EOX)  {			//end of SysEx
            if (guTxPackCnt == 1)  {
//PACKTYPE=0x05: send a 1 byte payload (=EOX) packet
                guPACKTYPE = 5;
                guMSTATUS = EOX;
                guMIDIDAT1 = 0;
                guMIDIDAT2 = 0;
                guMSTATUSok = 0;  //no running state with 1 byte messages
                guMDATA1ok = 0;
                guMDATA2ok = 0;

                if (guSYSEXBODY)   {
                    guMSTATUS = EOX;
                    guMIDIDAT1 = 0;
                    guMIDIDAT2 = 0;
                    guSYSEXBODY = 0;
                    goto SEND_AND_CLEANUP;
                }
                else   {
                    guMSTATUSok = 0;        //undefined MIDI stream
                    guTxPackCnt = 0;        //inore, no error message
                    guMDATA1ok = 0;
                    guMDATA2ok = 0;
                    guSYSEXBODY = 0;
                    return;
                }
            }
            else if (guTxPackCnt == 2)  {   //EOX is last of 2 byte packet
//PACKTYPE=0x06:SysEx ends with 2 bytes. Send 2 byte payload packet
                guPACKTYPE = 6;
                if (guMSTATUS != 0xF0) guMSTATUS = guMIDIDAT1;
                                //if (guMSTATUS==0xF0) incomplete SYSEX Msg
                guMSTATUS = guMIDIDAT1;   //else reorganize
                guMIDIDAT1 = EOX;
                guMIDIDAT2= 0;
                guMSTATUSok = 0;  //no running state with SysEx
                guMDATA1ok = 0;
                guMDATA2ok = 0;
                guSYSEXBODY = 0;

                goto SEND_AND_CLEANUP;

            }
            else   {     //EOX is last of 3 byte packet
//PACKTYPE=0x07: SysEx ends with 3 bytes. Send 3 byte payload packet
                guPACKTYPE = 7;
                if ((guMSTATUS==0xF0) && (guSYSEXBODY==0)) guMIDIDAT2 = EOX;
                                                            //tiny SysEx Msg
                else   {
                    guMSTATUS = guMIDIDAT1;   //reorganize
                    guMIDIDAT1 = guMIDIDAT2;
                    guMIDIDAT2 = EOX;
                }
                guMSTATUSok = 0;  //no running state with SysEx
                guMDATA1ok = 0;
                guMDATA2ok = 0;
                guSYSEXBODY = 0;

                goto SEND_AND_CLEANUP;

            }
        }
//handle different status bytes
        guMSTATUS = recv;
        if (guMSTATUS < 0xF4) { //1 or 2 data bytes Channel- or SystemCommon Msg
	        guMSTATUSok = 0xFF;
            guMDATA1ok = 0;
            guMDATA2ok = 0;
            guSYSEXBODY = 0;
            return;             //wait for subsequent data byte(s)
        }
        if ((guMSTATUS==0xF4) || (guMSTATUS==0xF5) ||
									(guMSTATUS==0xF9) ||(guMSTATUS==0xFD)) {
										//these message types are undefined
 	        guMSTATUSok = 0;
            guMDATA1ok = 0;
            guMDATA2ok = 0;
            guSYSEXBODY = 0;
            return;
        }
        else  {
//handle remaining 1 byte messages:
//0xF6 = Tune Request
//0xF8 = Timing Clock
//0xFA = MIDI Start
//0xFB = MIDI Continue
//0xFC = MIDI Stop
//0xFE = Active Sensing
//0xFF = MIDI Reset
//PACKTYPE=0x05: send a 1 byte payload packet
            guPACKTYPE = 5;
            guMIDIDAT1 = 0;
            guMIDIDAT2 = 0;
 	        guMSTATUSok = 0;
            guMDATA1ok = 0;
            guMDATA2ok = 0;
            guSYSEXBODY = 0;

            goto SEND_AND_CLEANUP;

        }
    }
    else  {   //handle data bytes (< 0x80)
        msgType = guMSTATUS & 0xF0;
        if (guMSTATUSok == 0)  {    //undefined MIDI stream
            guTxPackCnt = 0;        //inore, no error message
            guMDATA1ok = 0;
            guMDATA2ok = 0;
            guSYSEXBODY = 0;
            return;
        }
        if (guMDATA1ok == 0)   {    //first data byte
            guMIDIDAT1 = recv;      //keep data byte 1 in mind
            guMIDIDAT2 = 0;		    //preemptive for messages with 1 data byte

            if ((guMSTATUS == 0xF1) || (guMSTATUS == 0xF3))  	{
                //2 byte message Timecode or 2 byte message Song Select ready
                //PACKTYPE=0x02: send 2 byte payload packet
                guPACKTYPE = 2;
                guMSTATUSok = 0;    //no Running State with System Common
                guMDATA1ok = 0;
                guMDATA2ok = 0;
                guSYSEXBODY = 0;

                goto SEND_AND_CLEANUP;

            }
            if ((msgType == 0xC0) || (msgType == 0xD0))  {
                //PACKTYPE=0x0C or 0x0D: Channel Msg: send 2 byte payload packet
                guPACKTYPE = (msgType>>4) & 0x0F;
                guMSTATUSok	= 0xFF;     //confirm Running State for next packet
                guMDATA1ok = 0;
                guMDATA2ok = 0;
                guSYSEXBODY = 0;

                goto SEND_AND_CLEANUP;
            }
            else {
                guMDATA1ok = 0xFF;
                return;                 //wait for data byte 2
            }
        }
        else   {  //data byte 2 found
            if (guMSTATUSok == 0) {		//integry check
                guTxPackCnt = 0;        //inore, no error message
                guMDATA1ok = 0;
                guMDATA2ok = 0;
                guSYSEXBODY = 0;
                return;
            }
            if (guMDATA2ok == 0)  {
                guMIDIDAT2 = recv;
                if (guSYSEXBODY)  {
                    guMDATA2ok = 0xFF;
                    return;			    //SysEx body: wait for last byte of packet
                }

                if (guMSTATUS == 0xF2) {  //3 byte message Song Position Pointer ready
                                          // PACKTYPE=0x03: send 3 byte payload packet
                    guPACKTYPE = 3;
                    guMSTATUSok = 0;      //no running state with System Common
                    guMDATA1ok = 0;
                    guMDATA2ok = 0;
                    guSYSEXBODY = 0;

                    goto SEND_AND_CLEANUP;

                }
                else if (guMSTATUS == 0xF0)  {   //first 3 bytes of SysEx message ready
                                        //PACKTYPE=0x04: send 3 byte payload packet
                    guPACKTYPE = 4;
                    guSYSEXBODY = 0xFF; //subsequent data bytes are part of SysEx message
                    guMSTATUSok = 0xFF; //confirm for next SysEx packet
                    guMDATA1ok = 0;
                    guMDATA2ok = 0;

                    goto SEND_AND_CLEANUP;

                }
                else  { //PACKTYPE=0x08,0x09,0x0A,0x0B,0x0E: send 3 byte payload packet
                    guPACKTYPE = (msgType>>4) & 0x0F;
                    guMSTATUSok = 0xFF;  //confirm Running State for next packet
                    guMDATA1ok = 0;
                    guMDATA2ok = 0;

                    goto SEND_AND_CLEANUP;

                }
            }
            else  {   //3rd data byte within a SysEx message
//special case: data byte within a SysEx message
//reorganize to get a standard packet
                guMSTATUS = guMIDIDAT1;
                guMIDIDAT1 = guMIDIDAT2;
                guMIDIDAT2 = recv;
                guPACKTYPE = 4;         //PACKTYPE=0x04: send 3 byte payload packet
                guSYSEXBODY = 0xFF;     //subsequent data bytes are part of SysEx message
                guMSTATUSok = 0xFF;     //confirm for next pack
                guMDATA1ok = 0;
                guMDATA2ok = 0;
            }
        }
    }

SEND_AND_CLEANUP:

/*
	sendUsbByte(guPACKTYPE);
	sendUsbByte(guMSTATUS);
	sendUsbByte(guMIDIDAT1);
	sendUsbByte(guMIDIDAT2);
*/

    if (guConfigured == 0)  return;
// make sure that the buffer header is always incremented by a complete MIDI packet
// because the USB interrupt is asynchronous with this application level routine
// and blocking of the USB interrupt shall be avoided
    tempHead = gUSB_TXHEAD;
    *tempHead++ = guPACKTYPE;
    if (tempHead == gUSB_TXBUFTOP)  tempHead = gUSB_TXBUF;
     *tempHead++ = guMSTATUS;
    if (tempHead == gUSB_TXBUFTOP)  tempHead = gUSB_TXBUF;
    *tempHead++ = guMIDIDAT1;
    if (tempHead == gUSB_TXBUFTOP)  tempHead = gUSB_TXBUF;
    *tempHead++ = guMIDIDAT2;
    if (tempHead == gUSB_TXBUFTOP)  tempHead = gUSB_TXBUF;
    gUSB_TXHEAD = tempHead;

    guTxPackCnt = 0;      //ready for next package
    guMDATA1ok = 0;
    guMDATA2ok = 0;
    if (((guPACKTYPE<8) && (guPACKTYPE!=4)) || (guMSTATUS > 0xF0)) guMSTATUSok = 0;
            // else guMSTATUSok and guMSTATUS have to be preserved for Running State
            // guSYSEXBODY can only be reset by EOX or non SysEx MIDI message
}


//##########################################################################

